@@ -39,7 +39,7 @@
 #include "yelp-application.h"
 #include "yelp-window.h"
 
-#define DEFAULT_URI "ghelp:user-guide"
+#define DEFAULT_URI "ghelp:gnome-help"
 
 static gboolean editor_mode = FALSE;
 
@@ -707,8 +707,8 @@ yelp_application_is_bookmarked (YelpBookmarks     *bookmarks,
     GSettings *settings;
     YelpApplication *app = YELP_APPLICATION (bookmarks);
 
-    g_return_if_fail (page_id);
-    g_return_if_fail (doc_uri);
+    g_return_val_if_fail (page_id, FALSE);
+    g_return_val_if_fail (doc_uri, FALSE);
 
     settings = application_get_doc_settings (app, doc_uri);
     if (settings == NULL)
@@ -783,33 +783,6 @@ yelp_application_get_bookmarks (YelpApplication *app,
     return g_settings_get_value (settings, "bookmarks");
 }
 
-static void
-packages_installed (GDBusConnection *connection,
-                    GAsyncResult *res,
-                    YelpApplication *app)
-{
-    GError *error = NULL;
-    g_dbus_connection_call_finish (connection, res, &error);
-    if (error) {
-        const gchar *err = NULL;
-        if (error->domain == G_DBUS_ERROR) {
-            if (error->code == G_DBUS_ERROR_SERVICE_UNKNOWN)
-                err = _("You do not have PackageKit installed. Package installation links require PackageKit.");
-            else
-                err = error->message;
-        }
-        if (err != NULL) {
-            GtkWidget *dialog = gtk_message_dialog_new (NULL, 0,
-                                                        GTK_MESSAGE_ERROR,
-                                                        GTK_BUTTONS_CLOSE,
-                                                        "%s", err);
-            gtk_dialog_run ((GtkDialog *) dialog);
-            gtk_widget_destroy (dialog);
-        }
-        g_error_free (error);
-    }
-}
-
 void
 yelp_application_add_read_later (YelpApplication   *app,
                                  const gchar       *doc_uri,
@@ -880,30 +853,6 @@ yelp_application_get_read_later (YelpApplication *app,
     return g_settings_get_value (settings, "readlater");
 }
 
-void
-yelp_application_install_package (YelpApplication  *app,
-                                  const gchar      *pkg,
-                                  const gchar      *alt)
-{
-    GVariantBuilder *strv;
-    YelpApplicationPrivate *priv = GET_PRIV (app);
-    guint32 xid = 0;
-    strv = g_variant_builder_new (G_VARIANT_TYPE ("as"));
-    g_variant_builder_add (strv, "s", pkg);
-    g_dbus_connection_call (priv->connection,
-                            "org.freedesktop.PackageKit",
-                            "/org/freedesktop/PackageKit",
-                            "org.freedesktop.PackageKit.Modify",
-                            "InstallPackageNames",
-                            g_variant_new ("(uass)", xid, strv, ""),
-                            NULL,
-                            G_DBUS_CALL_FLAGS_NONE,
-                            -1, NULL,
-                            (GAsyncReadyCallback) packages_installed,
-                            app);
-    g_variant_builder_unref (strv);
-}
-
 static void
 bookmarks_changed (GSettings       *settings,
                    const gchar     *key,
